#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import random

from collections.abc import Iterable
from utils.types import Solution, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move


# ------------------------------- Neighbourhood ------------------------------

class PopNeighbourhood(Neighbourhood):
    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        yield PopMove(self)


# ----------------------------------- Move -----------------------------------

class PopMove(Move):
    def __init__(self, neighbourhood, increment=None):
        self.neighbourhood = neighbourhood
        self.increment = None
        self._cached_hash = hash(1)

    def __eq__(self, other):
        if isinstance(other, PopMove):
            return True
        else:
            return False        

    def __hash__(self):
        return self._cached_hash

    def invert_move(self) -> Move:
        raise NotImplementedError("This move cannot be inverted")

    def __str__(self):
        return "pop: %d, incr: %d" % (self.sequence[-1], self.increment)

    def apply(self, solution: Solution) -> Solution:
        #Update completion time
        index = len(solution.sequence)
        solution.completion_time[index-1] = 0
        solution.tardiness[index-1] = 0
        # Update lower bound
        solution.lb += self.increment if self.increment is not None else 0
        # Update value
        if len(solution.not_scheduled)!=0:
            solution.value = None
        else:
            solution.value = solution.lb
        # Update solution
        solution.sequence.remove(self.job)
        solution.not_scheduled.add(self.job)

    def lower_bound_increment(self, solution: Solution) -> Increment:
        prob = solution.problem
        index = len(solution.sequence)
        self.increment = -prob.weights[self.job] * solution.tardiness[index-1]
        return self.increment
    