#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import random

from collections.abc import Iterable
from utils.types import Solution, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move

# ------------------------------- Neighbourhood ------------------------------

class RemoveNeighbourhood(Neighbourhood):
    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        for job in solution.sequence:
            # i is the job to remove
            yield RemoveMove(self, job)

    def random_moves_without_replacement(self, solution: Solutiuon) -> Iterable[Move]:
        assert self.problem == solution.problem
        for job in random.sample(solution.sequence, k=len(solution.sequence)):
            yield RemoveMove(self, job)





# ----------------------------------- Move -----------------------------------

class RemoveMove(Move):
    def __init__(self, neighbourhood, job, increment=None):
        self.neighbourhood = neighbourhood
        # i is the job to remove
        self.job = job
        self.increment = None
        self._cached_hash = hash(self.job)

    def __str__(self):
        return "remove: %d, incr: %d" % (self.i, self.increment)
        
    def __eq__(self, other):
        if not isinstance(other, RemoveMove):
            return False
        return self._cached_hash == other._cached_hash

    def __hash__(self):
        return self._cached_hash

    def invert_move(self) -> Move:
        return AddMove(self.neighbourhood, self.job, -self.increment) 

    def apply(self, solution: Solution) -> Solution:
        raise NotImplementedError("This method should be finished")
        #Update completion time
        index = len(solution.sequence)
        solution.completion_time[index-1] = 0
        solution.tardiness[index-1] = 0
        # Update lower bound
        solution.lb += self.increment if self.increment is not None else 0
        # Update value
        if len(solution.not_scheduled)!=0:
            solution.value = None
        else:
            solution.value = solution.lb
        # Update solution
        solution.sequence.remove(self.job)
        solution.not_scheduled.add(self.job)
        return solution

    def lower_bound_increment(self, solution: Solution) -> Increment:
        raise NotImplementedError("This method should be finished")

        prob = solution.problem
        index = len(solution.sequence)
        self.increment = -prob.weights[self.job] * solution.tardiness[index-1]
        return self.increment
    