
import math
import random


from collections.abc import Iterable
from utils.types import Solution, Problem, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move
from api.neighborhood_local_swap import SwapNeighbourhood, SwapMove
from api.neighborhood_local_insert import InsertNeighbourhood, InsertMove


class InsertOrSwap(Neighbourhood):
    def __init__(self, problem: Problem):
        self.problem = problem
        self._swap_nbhood = SwapNeighbourhood(problem)
        self._insert_nbhood = InsertNeighbourhood(problem)
        

    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            for ix in range(n-1):
                yield SwapMove(self._swap_nbhood, ix)

            for ix in range(n-1):
                for jx in range(n):
                   if ix!=jx:
                       yield InsertMove(self._insert_nbhood, ix, jx)

    def random_move(self, solution: Solution) -> Move | None:
        """
        If the solution is feasible, return a random local move that can be
        applied to it.
        Note: repeated calls to this method may return the same
        local move.
        """
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            ix=random.randint(0,n-2)
            if random.random() < 0.5:
                # Swap moves
                return SwapMove(self._swap_nbhood, ix)
            else:
                # Insert moves
                jx = random.randint(0,n-1)
                while ix==jx:
                    jx = random.randint(0,n-1)
                if random.random() < 0.5:
                    return InsertMove(self._insert_nbhood, ix, jx)
                else:
                    return InsertMove(self._insert_nbhood, jx, ix)
        raise ValueError("Solution is not feasible, cannot generate random move.")

    def random_moves_without_replacement(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            for ix in random.sample(range(n-1), k=n-1):
                yield SwapMove(self._swap_nbhood, ix)
            for ix in random.sample(range(n-1), k=n-1):
                for jx in random.sample(range(n), k=n-1):
                    if ix!=jx:
                        yield InsertMove(self._insert_nbhood, ix, jx)