#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import math
import random

Value = int | float

Increment = int | float

class Problem:
    def __init__(self, processing_times, due_dates, weights):
        self.n = len(processing_times)
        self.processing_times = processing_times
        self.due_dates = due_dates
        self.weights = weights

    def __str__(self):
        p = " ".join(list(map(str, self.processing_times)))
        d = " ".join(list(map(str, self.due_dates)))
        w = " ".join(list(map(str, self.weights)))
        return f"Problem with {self.n} jobs\nProcessing times:\t {p}\nWeights:\t {w}\nDue dates:\t {d}"

    @classmethod
    def from_textio(cls, f):
        """
        Create a problem from a text I/O source `f`
        """
        processing_times_str = f.readline()
        weights_str = f.readline()
        due_dates_str = f.readline()
        processing_times = list(map(int, processing_times_str.split()))
        weights = list(map(int, weights_str.split()))
        due_dates = list(map(int, due_dates_str.split()))
        n = len(processing_times)
        if n != len(weights) or n != len(due_dates):
            raise ValueError("Processing times, weights and due dates must have the same length")
        return cls(processing_times, due_dates, weights)



