
from __future__ import annotations
import random, operator, time, math
import logging
logger = logging.getLogger("solvers")

from utils.types import Solution, Value

def best_improvement(solution: Solution):
    # modifies solution in place and returns a reference to it
    local_nbhood = solution.local_neighbourhood()
    move_iter = iter(local_nbhood.moves(solution))
    move = next(move_iter, None)
    while move is not None:
        best_move, best_incr = move, move.objective_value_increment(solution)
        for move in move_iter:
            incr = move.objective_value_increment(solution)
            if incr < best_incr:
                best_move, best_incr = move, incr
        if best_incr < 0:
            best_move.apply(solution)
            move_iter = iter(local_nbhood.moves(solution))
            logger.debug(f"new sol found: {best_incr}")
        move = next(move_iter, None)
    return solution

