#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>, Alexandre Jesus <me@adbjesus.com>
#
# SPDX-License-Identifier: Apache-2.0


import random, operator, time, math
import logging
logger = logging.getLogger("solvers")

from utils.types import Solution

def first_improvement(solution: Solution, local_nbhood=None):
    # modifies solution in place and returns a reference to it
    sol_value = solution.objective_value()
    if local_nbhood is None:     
        local_nbhood = solution.local_neighbourhood()
    move_iter = iter(local_nbhood.random_moves_without_replacement(solution))
    move = next(move_iter, None)
    while move is not None:
        increment = move.objective_value_increment(solution)
        if increment < 0:
            move.apply(solution)
            move_iter = iter(local_nbhood.random_moves_without_replacement(solution))
            sol_value += increment
            logger.debug(f"new sol found: {sol_value}")
        move = next(move_iter, None)
    return solution
