import random, operator, time, math


from time import perf_counter
from operator import itemgetter

from typing import cast

from utils.types import Solution, LocalSearch, Value, Increment


import logging
logger = logging.getLogger("solvers")

def grasp(solution: Solution,
          budget: float,
          alpha: float = 0.1,
          seed: int | None = None,
          local_search: LocalSearch[Solution] | None = None,
          ) -> Solution | None:
    if seed is not None:
        random.seed(seed)
    start = perf_counter()
    best, bestobj = None, None
    while perf_counter() - start < budget:
        s = solution.copy()
        b, bobj = (s.copy(), s.objective_value()) if s.is_feasible() else (None, None)
        move_iter = iter(s.construction_neighbourhood().moves(s))
        C = [(cast(Increment, c.lower_bound_increment(s)), c) for c in move_iter]
        while len(C) != 0:
            cmin = min(C, key=itemgetter(0))[0]
            cmax = max(C, key=itemgetter(0))[0]
            thresh = cmin + alpha * (cmax - cmin)
            RCL = [c for decr, c in C if decr <= thresh]
            c = random.choice(RCL)
            c.apply(s)
            if s.is_feasible():
                obj = cast(Value, s.objective_value())
                if bobj is None or obj < bobj:
                    b, bobj = s.copy(), obj
            move_iter = iter(s.construction_neighbourhood().moves(s))
            C = [(cast(Increment, c.lower_bound_increment(s)), c) for c in move_iter]
        if b is not None:
            if local_search is not None:
                b = local_search(b)
                if b is not None and b.is_feasible():
                    bobj = cast(Value, b.objective_value())
            if bestobj is None or bobj < bestobj:
                best, bestobj = b, bobj
                logger.debug(f"New best solution found: {bestobj}")
    return best
