#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>, Alexandre Jesus <me@adbjesus.com>
#
# SPDX-License-Identifier: Apache-2.0

import random, operator, math
import logging
logger = logging.getLogger("solvers")

from time import perf_counter

from utils.types import Solution

def greedy_construction(solution):
    constr_rule = solution.construction_neighbourhood()
    move_iter = iter(constr_rule.moves(solution))
    move = next(move_iter, None)
    while move is not None:
        best_move, best_incr = move, move.lower_bound_increment(solution)
        for move in move_iter:
            incr = move.lower_bound_increment(solution)
            if incr < best_incr:
                best_move, best_incr = move, incr
                if incr == 0:
                    break
        best_move.apply(solution)
        move_iter = iter(constr_rule.moves(solution))
        move = next(move_iter, None)
    return solution

def greedy_construction_random_tie_breaking(solution):
    constr_rule = solution.construction_neighbourhood()
    move_iter = iter(constr_rule.moves(solution))
    move = next(move_iter, None)
    while move is not None:
        best_move, best_incr = [move], move.lower_bound_increment(solution)
        for move in move_iter:
            incr = move.lower_bound_increment(solution)
            if incr == best_incr or type(incr) == float and math.isclose(best_incr, incr):
                best_move.append(move)
            elif incr < best_incr:
                best_move, best_incr = [move], incr
        random.choice(best_move).apply(solution)
        move_iter = iter(constr_rule.moves(solution))
        move = next(move_iter, None)
    return solution
