import random, operator, math
import time

import logging
logger = logging.getLogger("solvers")

from utils.types import Solution, Value, Neighbourhood, Move, Increment

def ils(solution: Solution, kickStrength: int, budget: int) -> Solution:
    start = time.perf_counter()
    local_nbhood = solution.local_neighbourhood()
    # FIXME: This is mainly for testing
    perturb_nbhood = solution.perturbation_neighbourhood()
    best = solution.copy()
    best_obj = best.objective_value()
    while time.perf_counter() - start < budget:
        for move in local_nbhood.random_moves_without_replacement(solution):
            incr = move.objective_value_increment(solution)
            if incr < 0:
                move.apply(solution)
                logger.debug(f"new sol found: {incr} {solution.objective_value()}")
                break
            if time.perf_counter() - start >= budget:
                if solution.objective_value() < best_obj:
                    return solution
                else:
                    return best
        else:
            # Local optimum found
            obj = solution.objective_value()
            if obj <= best_obj:
                best = solution.copy()
                best_obj = obj
            else:
                solution = best.copy()
            # Very simple perturbation
            for _ in range(kickStrength):
                perturb_nbhood.random_move(solution).apply(solution)
            logger.debug(f"Perturbation applied: {solution.value}")
        # Check if we are still within the budget
    if solution.value < best_obj:
        return solution
    else:
        return best
