
import random, operator, time, math
import logging
logger = logging.getLogger("solvers")

from utils.types import Solution

from roas.first_improvement import first_improvement
from roas.best_improvement import best_improvement

from time import perf_counter

def basic_variable_neighbourhood_descent(solution: Solution):
    # modifies solution in place and returns a reference to it
    # sol_value = solution.objective_value()
    local_nbhoods = solution.local_sub_neighbourhoods()
    n_local_nbhoods = len(local_nbhoods)
    if n_local_nbhoods == 0:
        return solution
    k=0
    while k < n_local_nbhoods:
        local_nbhood = local_nbhoods[k]
        move_iter = iter(local_nbhood.moves(solution))
        move = next(move_iter, None)        
        best_move, best_incr = move, move.objective_value_increment(solution)
        for move in move_iter:
            incr = move.objective_value_increment(solution)
            if incr < best_incr:
                best_move, best_incr = move, incr
        if best_incr < 0:
            best_move.apply(solution)
            #sol_value += best_incr
            k=0
            logger.debug(f"new sol found: {solution.objective_value()}")
        else:
            k += 1 
    return solution

def variable_neighbourhood_descent(solution: Solution):
    # modifies solution in place and returns a reference to it
    # sol_value = solution.objective_value()
    local_nbhoods = solution.local_sub_neighbourhoods()
    n_local_nbhoods = len(local_nbhoods)
    if n_local_nbhoods == 0:
        return solution
    k=0
    while k < n_local_nbhoods:
        local_nbhood = local_nbhoods[k]
        new_solution = first_improvement(solution)
        if new_solution.value < solution.value:
            solution = new_solution
            k=0
            logger.debug(f"new sol found: {solution.value}")
        else:
            k += 1
    return solution


def basic_variable_neighbourhood_search(solution: Solution, time_budget: float = math.inf, max_no_improv: int = math.inf):
    # modifies solution in place and returns a reference to it
    start_time = perf_counter()
    local_nbhoods = solution.local_sub_neighbourhoods()
    n_local_nbhoods = len(local_nbhoods)
    if n_local_nbhoods == 0:
        return solution

    best = solution.copy()
    best_obj = solution.value
    idle_iterations = 0
    while True:
        k=0
        while k < n_local_nbhoods:
            if perf_counter() - start_time > time_budget or idle_iterations > max_no_improv:
                return best
            local_nbhood = local_nbhoods[k]
            new_solution = solution.copy()
            move = local_nbhood.random_move(new_solution)
            move.apply(new_solution)
            new_solution = best_improvement(new_solution)
            idle_iterations += 1
            if new_solution.value < solution.value:
                solution = new_solution.copy()
                k=0
                if new_solution.value < best_obj:
                    best = new_solution.copy()
                    best_obj = solution.value
                    idle_iterations = 0
                    logger.debug(f"new best found: {solution.value}")
            else:
                k += 1
    return best

