
from __future__ import annotations

from typing import TypeVar, Protocol, Union, Any
from collections.abc import Callable, Iterable, Hashable
from collections import defaultdict



Value = TypeVar('Value', covariant=True)
#ComponentId = Hashable
Increment = TypeVar('Increment', bound=Value, covariant=True)


class MoveProtocol(Protocol):
    def apply(self, solution: Solution) -> None: ...
    def objective_value_increment(self, solution: Solution) -> Value: ...
    def lower_bound_increment(self, solution: Solution) -> Value: ...

Move = TypeVar('Move', bound=MoveProtocol)



class NeighbourhoodProtocol(Protocol[Move]): 
    def moves(self, solution: Solution) -> Iterable[Move]: ...
    def random_move(self, solution: Solution) -> Move | None: ...
    def random_moves_without_replacement(self, solution: Solution) -> Iterable[Move]: ...   
    def invert_move(self, move: Move) -> Move: ...

Neighbourhood = TypeVar('Neighbourhood', bound=NeighbourhoodProtocol)



class ProblemProtocol(Protocol):
    ...

Problem = TypeVar('Problem', bound=ProblemProtocol)

#class ComponentProtocol(Protocol):
#    @property
#    def cid(self) -> ComponentId: ...

# Component = TypeVar('Component', bound=ComponentProtocol)

class SolutionProtocol(Protocol[Value, Move, Neighbourhood]):
    def problem_instance(self) -> Problem: ...
    def lower_bound(self) -> Value | None: ...
    def objective_value(self) -> Value | None: ...
    def copy(self: Solution) -> Solution: ...
    def is_feasible(self) -> bool: ...
    def local_sub_neighbourhoods(self) -> list[Neighbourhood]: ...
    def local_neighbourhood(self) -> Neighbourhood: ...
    def construction_neighbourhood(self) -> Neighbourhood: ...
    
Solution = TypeVar('Solution', bound=SolutionProtocol)




class LocalSearch(Protocol[Solution]):
    def __call__(self, s: Solution) -> Solution: ...